@@ -1,4 +1,4 @@
-// Copyright 2019 Parity Technologies (UK) Ltd.
+// Copyright 2019-2021 Parity Technologies (UK) Ltd.
 //
 // Permission is hereby granted, free of charge, to any
 // person obtaining a copy of this software and associated
@@ -32,11 +32,14 @@
 
 extern crate alloc;
 
+mod future;
 mod server;
 
 #[cfg(test)]
 mod tests;
 
+pub use future::{ServerHandle as WsServerHandle, ShutdownWaiter as WsShutdownWaiter};
 pub use jsonrpsee_types as types;
 pub use jsonrpsee_utils::server::rpc_module::{RpcModule, SubscriptionSink};
-pub use server::{Builder as WsServerBuilder, Server as WsServer, StopHandle as WsStopHandle};
+pub use server::{Builder as WsServerBuilder, Server as WsServer};
+pub use tracing;
