@@ -32,36 +32,41 @@ use std::task::{Context, Poll};
 use crate::future::{FutureDriver, ServerHandle, StopMonitor};
 use crate::types::{
 	error::Error,
+	middleware::Middleware,
 	v2::{ErrorCode, Id, Request},
 	TEN_MB_SIZE_BYTES,
 };
 use futures_channel::mpsc;
+use futures_util::future::join_all;
 use futures_util::future::FutureExt;
 use futures_util::io::{BufReader, BufWriter};
-use futures_util::stream::{self, StreamExt};
+use futures_util::stream::StreamExt;
 use soketto::connection::Error as SokettoError;
 use soketto::handshake::{server::Response, Server as SokettoServer};
 use soketto::Sender;
 use tokio::net::{TcpListener, TcpStream, ToSocketAddrs};
 use tokio_util::compat::{Compat, TokioAsyncReadCompatExt};
 
-use jsonrpsee_utils::server::helpers::{collect_batch_response, prepare_error, send_error};
-use jsonrpsee_utils::server::resource_limiting::Resources;
-use jsonrpsee_utils::server::rpc_module::{ConnectionId, Methods};
+use jsonrpsee_utils::server::{
+	helpers::{collect_batch_response, prepare_error, MethodSink},
+	resource_limiting::Resources,
+	rpc_module::{ConnectionId, MethodResult, Methods},
+};
 
 /// Default maximum connections allowed.
 const MAX_CONNECTIONS: u64 = 100;
 
 /// A WebSocket JSON RPC server.
 #[derive(Debug)]
-pub struct Server {
+pub struct Server<M> {
 	listener: TcpListener,
 	cfg: Settings,
 	stop_monitor: StopMonitor,
 	resources: Resources,
+	middleware: M,
 }
 
-impl Server {
+impl<M: Middleware> Server<M> {
 	/// Returns socket address to which the server is bound.
 	pub fn local_addr(&self) -> Result<SocketAddr, Error> {
 		self.listener.local_addr().map_err(Into::into)
@@ -88,6 +93,7 @@ impl Server {
 	async fn start_inner(self, methods: Methods) {
 		let stop_monitor = self.stop_monitor;
 		let resources = self.resources;
+		let middleware = self.middleware;
 
 		let mut id = 0;
 		let mut connections = FutureDriver::default();
@@ -118,6 +124,7 @@ impl Server {
 							resources: &resources,
 							cfg,
 							stop_monitor: &stop_monitor,
+							middleware: middleware.clone(),
 						},
 					)));
 
@@ -184,7 +191,7 @@ where
 	}
 }
 
-enum HandshakeResponse<'a> {
+enum HandshakeResponse<'a, M> {
 	Reject {
 		status_code: u16,
 	},
@@ -194,10 +201,14 @@ enum HandshakeResponse<'a> {
 		resources: &'a Resources,
 		cfg: &'a Settings,
 		stop_monitor: &'a StopMonitor,
+		middleware: M,
 	},
 }
 
-async fn handshake(socket: tokio::net::TcpStream, mode: HandshakeResponse<'_>) -> Result<(), Error> {
+async fn handshake<M>(socket: tokio::net::TcpStream, mode: HandshakeResponse<'_, M>) -> Result<(), Error>
+where
+	M: Middleware,
+{
 	// For each incoming background_task we perform a handshake.
 	let mut server = SokettoServer::new(BufReader::new(BufWriter::new(socket.compat())));
 
@@ -214,7 +225,7 @@ async fn handshake(socket: tokio::net::TcpStream, mode: HandshakeResponse<'_>) -
 
 			Ok(())
 		}
-		HandshakeResponse::Accept { conn_id, methods, resources, cfg, stop_monitor } => {
+		HandshakeResponse::Accept { conn_id, methods, resources, cfg, stop_monitor, middleware } => {
 			tracing::debug!("Accepting new connection: {}", conn_id);
 			let key = {
 				let req = server.receive_request().await?;
@@ -244,6 +255,7 @@ async fn handshake(socket: tokio::net::TcpStream, mode: HandshakeResponse<'_>) -
 				resources.clone(),
 				cfg.max_request_body_size,
 				stop_monitor.clone(),
+				middleware,
 			))
 			.await;
 
@@ -262,6 +274,7 @@ async fn background_task(
 	resources: Resources,
 	max_request_body_size: u32,
 	stop_server: StopMonitor,
+	middleware: impl Middleware,
 ) -> Result<(), Error> {
 	// And we can finally transition to a websocket background_task.
 	let mut builder = server.into_builder();
@@ -269,6 +282,9 @@ async fn background_task(
 	let (mut sender, mut receiver) = builder.finish();
 	let (tx, mut rx) = mpsc::unbounded::<String>();
 	let stop_server2 = stop_server.clone();
+	let sink = MethodSink::new_with_limit(tx, max_request_body_size);
+
+	middleware.on_connect();
 
 	// Send results back to the client.
 	tokio::spawn(async move {
@@ -293,8 +309,9 @@ async fn background_task(
 	// Buffer for incoming data.
 	let mut data = Vec::with_capacity(100);
 	let mut method_executors = FutureDriver::default();
+	let middleware = &middleware;
 
-	loop {
+	let result = loop {
 		data.clear();
 
 		{
@@ -307,8 +324,8 @@ async fn background_task(
 				match err {
 					MonitoredError::Selector(SokettoError::Closed) => {
 						tracing::debug!("WS transport error: remote peer terminated the connection: {}", conn_id);
-						tx.close_channel();
-						return Ok(());
+						sink.close();
+						break Ok(());
 					}
 					MonitoredError::Selector(SokettoError::MessageTooLarge { current, maximum }) => {
 						tracing::warn!(
@@ -316,35 +333,56 @@ async fn background_task(
 							current,
 							maximum
 						);
-						send_error(Id::Null, &tx, ErrorCode::OversizedRequest.into());
+						sink.send_error(Id::Null, ErrorCode::OversizedRequest.into());
 						continue;
 					}
 					// These errors can not be gracefully handled, so just log them and terminate the connection.
 					MonitoredError::Selector(err) => {
 						tracing::error!("WS transport error: {:?} => terminating connection {}", err, conn_id);
-						tx.close_channel();
-						return Err(err.into());
+						sink.close();
+						break Err(err.into());
 					}
-					MonitoredError::Shutdown => break,
+					MonitoredError::Shutdown => break Ok(()),
 				};
 			};
 		};
 
 		tracing::debug!("recv {} bytes", data.len());
 
+		let request_start = middleware.on_request();
+
 		match data.get(0) {
 			Some(b'{') => {
 				if let Ok(req) = serde_json::from_slice::<Request>(&data) {
+					middleware.on_call(req.method.as_ref());
+
 					tracing::debug!("recv method call={}", req.method);
 					tracing::trace!("recv: req={:?}", req);
-					if let Some(fut) =
-						methods.execute_with_resources(&tx, req, conn_id, &resources, max_request_body_size)
-					{
-						method_executors.add(fut);
+					match methods.execute_with_resources(&sink, req, conn_id, &resources) {
+						Ok((name, MethodResult::Sync(success))) => {
+							middleware.on_result(name, success, request_start);
+							middleware.on_response(request_start);
+						}
+						Ok((name, MethodResult::Async(fut))) => {
+							let request_start = request_start;
+
+							let fut = async move {
+								let success = fut.await;
+								middleware.on_result(name, success, request_start);
+								middleware.on_response(request_start);
+							};
+
+							method_executors.add(fut.boxed());
+						}
+						Err(name) => {
+							middleware.on_result(name.as_ref(), false, request_start);
+							middleware.on_response(request_start);
+						}
 					}
 				} else {
 					let (id, code) = prepare_error(&data);
-					send_error(id, &tx, code.into());
+					sink.send_error(id, code.into());
+					middleware.on_response(request_start);
 				}
 			}
 			Some(b'[') => {
@@ -352,57 +390,71 @@ async fn background_task(
 				let d = std::mem::take(&mut data);
 				let resources = &resources;
 				let methods = &methods;
-				let tx2 = tx.clone();
+				let sink = sink.clone();
 
 				let fut = async move {
 					// Batch responses must be sent back as a single message so we read the results from each
 					// request in the batch and read the results off of a new channel, `rx_batch`, and then send the
 					// complete batch response back to the client over `tx`.
 					let (tx_batch, mut rx_batch) = mpsc::unbounded();
+					let sink_batch = MethodSink::new_with_limit(tx_batch, max_request_body_size);
 					if let Ok(batch) = serde_json::from_slice::<Vec<Request>>(&d) {
 						tracing::debug!("recv batch len={}", batch.len());
 						tracing::trace!("recv: batch={:?}", batch);
 						if !batch.is_empty() {
-							let methods_stream = stream::iter(batch.into_iter().filter_map(|req| {
-								methods.execute_with_resources(
-									&tx_batch,
-									req,
-									conn_id,
-									resources,
-									max_request_body_size,
-								)
-							}));
+							join_all(batch.into_iter().filter_map(move |req| {
+								match methods.execute_with_resources(&sink_batch, req, conn_id, resources) {
+									Ok((name, MethodResult::Sync(success))) => {
+										middleware.on_result(name, success, request_start);
+										None
+									}
+									Ok((name, MethodResult::Async(fut))) => Some(async move {
+										let success = fut.await;
+										middleware.on_result(name, success, request_start);
+									}),
+									Err(name) => {
+										middleware.on_result(name.as_ref(), false, request_start);
+										None
+									}
+								}
+							}))
+							.await;
 
-							let results = methods_stream
-								.for_each_concurrent(None, |item| item)
-								.then(|_| {
-									rx_batch.close();
-									collect_batch_response(rx_batch)
-								})
-								.await;
+							rx_batch.close();
+							let results = collect_batch_response(rx_batch).await;
 
-							if let Err(err) = tx2.unbounded_send(results) {
+							if let Err(err) = sink.send_raw(results) {
 								tracing::error!("Error sending batch response to the client: {:?}", err)
+							} else {
+								middleware.on_response(request_start);
 							}
 						} else {
-							send_error(Id::Null, &tx2, ErrorCode::InvalidRequest.into());
+							sink.send_error(Id::Null, ErrorCode::InvalidRequest.into());
+							middleware.on_response(request_start);
 						}
 					} else {
 						let (id, code) = prepare_error(&d);
-						send_error(id, &tx2, code.into());
+						sink.send_error(id, code.into());
+						middleware.on_response(request_start);
 					}
 				};
 
 				method_executors.add(Box::pin(fut));
 			}
-			_ => send_error(Id::Null, &tx, ErrorCode::ParseError.into()),
+			_ => {
+				sink.send_error(Id::Null, ErrorCode::ParseError.into());
+			}
 		}
-	}
+	};
 
-	// Drive all running methods to completion
+	middleware.on_disconnect();
+
+	// Drive all running methods to completion.
+	// **NOTE** Do not return early in this function. This `await` needs to run to guarantee
+	// proper drop behaviour.
 	method_executors.await;
 
-	Ok(())
+	result
 }
 
 #[derive(Debug, Clone)]
@@ -453,13 +505,27 @@ impl Default for Settings {
 }
 
 /// Builder to configure and create a JSON-RPC Websocket server
-#[derive(Debug, Default)]
-pub struct Builder {
+#[derive(Debug)]
+pub struct Builder<M = ()> {
 	settings: Settings,
 	resources: Resources,
+	middleware: M,
+}
+
+impl Default for Builder {
+	fn default() -> Self {
+		Builder { settings: Settings::default(), resources: Resources::default(), middleware: () }
+	}
 }
 
 impl Builder {
+	/// Create a default server builder.
+	pub fn new() -> Self {
+		Self::default()
+	}
+}
+
+impl<M> Builder<M> {
 	/// Set the maximum size of a request body in bytes. Default is 10 MiB.
 	pub fn max_request_body_size(mut self, size: u32) -> Self {
 		self.settings.max_request_body_size = size;
@@ -511,6 +577,34 @@ impl Builder {
 		Ok(self)
 	}
 
+	/// Add a middleware to the builder [`Middleware`](../jsonrpsee_types/middleware/trait.Middleware.html).
+	///
+	/// ```
+	/// use jsonrpsee_types::middleware::Middleware;
+	/// use jsonrpsee_ws_server::WsServerBuilder;
+	/// use std::time::Instant;
+	///
+	/// #[derive(Clone)]
+	/// struct MyMiddleware;
+	///
+	/// impl Middleware for MyMiddleware {
+	///     type Instant = Instant;
+	///
+	///     fn on_request(&self) -> Instant {
+	///         Instant::now()
+	///     }
+	///
+	///     fn on_result(&self, name: &str, success: bool, started_at: Instant) {
+	///         println!("Call to '{}' took {:?}", name, started_at.elapsed());
+	///     }
+	/// }
+	///
+	/// let builder = WsServerBuilder::new().set_middleware(MyMiddleware);
+	/// ```
+	pub fn set_middleware<T: Middleware>(self, middleware: T) -> Builder<T> {
+		Builder { settings: self.settings, resources: self.resources, middleware }
+	}
+
 	/// Restores the default behavior of allowing connections with `Origin` header
 	/// containing any value. This will undo any list set by [`set_allowed_origins`](Builder::set_allowed_origins).
 	pub fn allow_all_origins(mut self) -> Self {
@@ -578,11 +672,11 @@ impl Builder {
 	/// }
 	/// ```
 	///
-	pub async fn build(self, addrs: impl ToSocketAddrs) -> Result<Server, Error> {
+	pub async fn build(self, addrs: impl ToSocketAddrs) -> Result<Server<M>, Error> {
 		let listener = TcpListener::bind(addrs).await?;
 		let stop_monitor = StopMonitor::new();
 		let resources = self.resources;
-		Ok(Server { listener, cfg: self.settings, stop_monitor, resources })
+		Ok(Server { listener, cfg: self.settings, stop_monitor, resources, middleware: self.middleware })
 	}
 }
 
